# General imports
import numpy as np
import scipy as scp 
import scipy.ndimage as ndi
import numpy.linalg as lna



def reconstruct_img(lmdata, iA):
    '''
    RECONSTRUCT_IMG Performs image reconstruction by wavelength using the inverted A-matrix.
    
    img = RECONSTRUCT_IMG(data, iA) takes the inverted VOX x MEAS
    sensitivity matrix "iA" and right-multiplies it by the logmeaned MEAS x
    TIME light-level matrix "lmdata" to reconstruct an image in voxel
    space. 
    
    The image is output in a VOX x TIME matrix "img".
    
    See Also: TIKHONOV_INVERT_AMAT, SMOOTH_AMAT, SPECTROSCOPY_IMG,
    FINDGOODMEAS.
    '''
    ## Parameters and Initialization
    units_scaling = 1/100 # Assuming OptProp in mm^-1 
    
    ## Reconstruct.
    cortex_mu_a = iA @ lmdata

    ## Correct units and convert to single.
    cortex_mu_a = np.single(np.multiply(cortex_mu_a, units_scaling))

    return cortex_mu_a

def smooth_Amat(iA_in, dim, gsigma):
    '''
    SMOOTH_AMAT Performs Gaussian smoothing on a sensitivity matrix.

    iA_out = SMOOTH_AMAT(iA_in, dim, gsigma) takes the inverted VOX x MEAS
    sensitivity matrix "iA_in" and performs Gaussian smoothing in the 3D
    voxel space on each of the concatenated measurement matrices within,
    returning it as "iA_out". The user specifies the Gaussian filter
    half-width "gsigma".

    See Also: TIKHONOV_INVERT_AMAT, RECONSTRUCT_IMG, FINDGOODMEAS.
    '''
    # Parameters and Initialization.
    Nvox = np.shape(iA_in)[0]
    Nm = np.shape(iA_in)[1]
    iA_out = np.zeros((Nvox, Nm))

    nVx =int(dim['nVx'])
    nVy = int(dim['nVy'])
    nVz = int(dim['nVz'])

    gsigma = gsigma / dim['sV']

    # Preallocate voxel space.
    if 'Good_Vox' in dim:
        GV = dim['Good_Vox'].astype(int)
    else:
        GV = np.ones((nVx, nVy, nVz)).astype(int) # WARNING: THIS RUNS WAY SLOWER.

    # Do smoothing in parallel.
    for k in range(0, Nm):
        iAvox = np.zeros((nVx, nVy, nVz)) # Set up temp iAvox
        iAvox = np.reshape(iAvox, (nVx * nVy * nVz, ), order = 'F')     # iAvox needs to be a vector for GV to be used as an index
        iAvox[GV-1] = iA_in[:, k]                                       
        iAvox = np.reshape(iAvox, (nVx, nVy, nVz), order = 'F')
        # Gaussian smoothing to replicate Matlab's imgaussfilt3()
        # imgaussfilt3's default filter size is 2*ceil(2*SIGMA)+1 so we set the truncate parameter to ceil(2*gsigma)/gsigma (=2, for input gsigma of 3)
        # imgaussfilt3 default pad info: "Input image values outside the bounds of the image are assumed equal to the nearest image border value." so we set mode = 'nearest'
        iAvox = ndi.gaussian_filter(iAvox, gsigma, mode = 'nearest', truncate=np.ceil(2*gsigma)/gsigma)
        iAvox= np.reshape(iAvox, (nVx * nVy * nVz, ), order = 'F')      # reshape iAvox like above or else GV cannot be used as index
        iA_out[:,k] = np.single(iAvox[GV-1])                            
        iAvox = np.reshape(iAvox, (nVx, nVy, nVz), order = 'F')
    
    return iA_out

def spectroscopy_img(cortex_mu_a, E):
    '''
    SPECTROSCOPY_IMG Completes the Beer-Lambert law from a reconstructed image.

    img_out = SPECTROSCOPY_IMG(img_in, E) takes the reconstructed VOX x
    TIME x WL mua image "img_in" and multiplies it by the inverse
    of the extinction coefficient matrix "E" to create an output VOX x TIME
    x HB matrix "img_out", where HB 1 and 2 are the voxel-space time series
    images for HbO and HbR, respectively.

    See Also: RECONSTRUCT_IMG, AFFINE3D_IMG.
    '''
    
    ## Parameters and Initialization.
    Nvox = cortex_mu_a.shape[0]
    Nt = cortex_mu_a.shape[1]
    Nc = cortex_mu_a.shape[2]
    E1 = E.shape[0]
    E2 = E.shape[1]
    umol_scale = 1000

    ## Check compatibility of the image and E matrix.
    if E1 != Nc:
        raise ValueError('Error: The image wavelengths and spectroscopy matrix dimensions do not match.')

    ## Invert Spectroscopy Matrix.
    iE = np.linalg.inv(E)

    #Initialize Outputs
    cortex_hb = np.zeros((Nvox, Nt, Nc), dtype = np.float64)
    for k in range(0, E2):
        temp = np.zeros((Nvox, Nt))
        for l in range(0, E1):
            temp = temp + np.multiply(np.squeeze(iE[k, l]), np.squeeze(cortex_mu_a[:,:,l]))
        cortex_hb[:, :, k] = temp

    cortex_hb = np.multiply(cortex_hb, umol_scale) # Fix units to umol

    return cortex_hb

def Tikhonov_invert_Amat(A, lambda1, lambda2):
    '''
    TIKHONOV_INVERT_AMAT Inverts a sensitivity "A" matrix.

    iA = TIKHONOV_INVERT_AMAT(A, lambda1, lambda2) allows the user to
    specify the values of the "lambda1" and "lambda2" parameters in the
    inversion calculation. lambda2 for spatially-variant regularization,
    is optional.

    See Also: SMOOTH_AMAT, RECONSTRUCT_IMG, FINDGOODMEAS.
    ''' 
    if not np.isreal(A.any):  # If complex A, sep into [Re;Im] first
        A = np.concatenate((A.re, A.imag), axis = 0)
    Nm = np.shape(A)[0]
    Nvox = np.shape(A)[1]
    if lambda2:
        svr = 1
    else:
        svr = 0

    ## Spatially variant regularization
    if svr:
        ll_0 = np.sum((A**2), axis = 0) 
        ll = np.sqrt(ll_0+lambda2*max(ll_0)) # Adjust with Lambda2 cut-off value
        A = A/ll

    ## Take the pseudo-inverse.
    if Nvox < Nm:
        Att = np.zeros(Nvox, dtype = np.single)
        Att = np.single(A.astype(np.transpose(A).astype(np.float64) @ np.float64))
        ss = lna.norm(Att, ord = 2) # numpy.linalg.norm() with ord = 2 is the matrix 2-norm 
        penalty = np.multiply(np.sqrt(ss), lambda1)
        iA = lna.lstsq(Att + np.multiply(penalty**2, np.eye(Nvox, dtype = np.uint8)), np.transpose(A))
    else:
        Att = np.zeros(Nm, dtype = np.single)
        Att = np.single(A.astype(np.float64) @ np.transpose(A).astype(np.float64))  
        ss = lna.norm(Att)
        penalty = np.multiply(np.sqrt(ss), lambda1)
        # In matlab, the following two lines are written as: iA = A' / (Att + penalty .^ 2 .* eye(Nm, 'single'));
        # In python, it is impossible to divide a (m,n) matrix by (n,n), when m or n != 1
        # Instead, we multiply A' by the inverse of the divisor 
        iAtt = lna.inv((Att + np.multiply(penalty**2, np.eye(Nm, dtype = np.uint8))))
        iA = np.matmul(np.transpose(A),iAtt)

    ## Undo spatially variant regularization
    if svr:
        ll.shape = (1, ll.shape[0]) # add singleton dimension to ll
        iA = iA / np.transpose(ll)

    return iA


# General imports
import math
import fractions
import numpy as np
import numpy.linalg as lna
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.colors as colors
import scipy.interpolate
import scipy as scp 
import numpy.matlib as mlb
import numpy.matlib as nm
import functools as ft
import scipy.signal as sig
import sympy as sym
import copy


from math import trunc
from pickle import NONE
from numpy import float64, matrix
from numpy.lib.shape_base import expand_dims
from matplotlib.pyplot import colorbar, colormaps
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from mpl_toolkits.mplot3d import Axes3D

import neuro_dot as ndot



def detrend_tts(data_in):
    """
    DETREND_TTS Performs linear detrending.

    data_out = DETREND_TTS(data_in) takes a raw light-level data array
    "data_in" of the format MEAS x TIME and removes the straight-line fit
    along the TIME dimension from each measurement, returning it as
    "data_out".

    See Also: LOGMEAN.
    """

    ## Parameters and initialization
    dims = np.shape(data_in)
    Nt = dims[-1]
    NDtf = np.ndim(data_in) > 2

    ## N-D Input.
    if NDtf:
        data_in = np.reshape(data_in, len(data_in)/Nt, Nt)

    ## Detrend.
    data_out = np.squeeze(sig.detrend(data_in[:,None])) 

    ## Remove mean
    if data_out.ndim == 1: # When data_in is a vector
        meanRow = np.ndarray.mean(data_out, dtype = np.float64)
        data_out= data_out - meanRow
    else:
        meanRow = np.ndarray.mean(data_out, axis = 1,dtype = np.float64)
        data_out= data_out - meanRow[:,None]

    ## N-D Output.
    if NDtf:
        data_out = np.reshape(data_out, dims)
    
    return data_out

def nextpow2(N):
    """ 
    NEXTPOW2 Finds the next power of 2, given an integer input.
    """
    n = 1
    while n < N: n = n^2
    return n

def fft_tts(data, framerate):
    """
    FFT_TTS Computes the Fourier Transform of a time-domain input.
    ftmag = FFT_TTS(data, framerate) takes a data array "data" of
    the MEAS x TIME format, pads it timewise to the next highest power of
    two (for better performance), performs the fast Fourier transform of
    each channel using the built-in MATLAB function FFT, normalizes by the
    padded time length, and takes the first half of the transformed data
    (which is the positive half of the frequency domain). The result is
    output into "ftmag".

    [ftmag, ftdomain] = FFT_TTS(data, framerate) also returns the
    corresponding normalized frequency domain "ftdomain", which extends
    from 0 to the Nyquist frequency, which is calculated from the input
    "framerate".

    [ftmag, ftdomain, ftpower] = FFT_TTS(data, framerate) also returns the
    power spectrum, which is the absolute value of the magnitude, squared.

    [ftmag, ftdomain, ftpower, ftphase] = FFT_TTS(data, framerate) also
    returns the phase at each frequency, as calculated by the MATLAB
    ANGLE function.

    Dependencies: NORMALIZE2RANGE_TTS.
    
    See Also: LOGMEAN, FFT, POW2, NEXTPOW2, ANGLE.
    """

    ## Parameters and Initialization.
    dims = data.shape
    Nt = dims[-1]
    NDtf = (len(dims) > 2)
    Ndft = 2 ** math.ceil(math.log2(abs(Nt))); # Zero pack to a power of 2.
    Nf = int(1 + Ndft / 2)
    
    # N-D Input.
    if NDtf:
        data = np.reshape(data, len(data)/Nt, Nt)

    # Remove mean.
    meanRow = np.ndarray.mean(data, axis = 1,dtype = np.float64)
    normdata = data - meanRow[:,None]

    # Prep data for FFT. 
    ftdomain = (framerate / 2) * np.linspace(0, 1, Nf) # domain of FFT: [zero:Nyquist]

    # Perform FFT.
    P = scp.fft.fft(normdata, Ndft, 1) / Ndft # Do FFT in time dimension and normalize by Ndft
    ftmag = math.sqrt(2) * abs(P[:, 0:Nf])    # Take positive frequencies, x2 for negative frequencies.

    # N-D Output.
    if NDtf:
        ftmag = np.reshape(ftmag, data.shape[0], Nf)

    # Other outputs.
    ftpower = abs(ftmag) ** 2
    ftphase = np.angle(P[:, 1:Nf])

    return ftdomain, ftmag, ftpower, ftphase

def gethem(data, info, sel_type  = 'r2d', value = [10,16]):
    """
    GETHEM Calculates the mean across a set of measurements.
    
    hem = GETHEM(data, info) takes a light-level array "data" of the format
    MEAS x TIME, and using the scan metadata in "info.pairs" averages the
    measurements for each wavelength present. The result is
    referred to as the "hem" of a measurement set. If there is a
    good measurements logical vector present in "info.MEAS.GI", it will be
    applied to the data; otherwise, "info.MEAS.GI" will be set to true for
    all measurements (i.e., all measurements are assumed to be good). The
    variable "hem" is output in the format WL x TIME.
    
    hem = GETHEM(data, info, sel_type, value) allows the user to set the
    criteria for determining shallow measurements. "sel_type" can be 'r2d',
    'r3d', or 'NN', corresponding to the columns of the "info.pairs" table,
    and "value" can either take the form of a two-element "[min, max]"
    vector (for 'r2d' and 'r3d'), or a scalar or vector containing all
    nearest neighbor numbers to be averaged. By default, this function
    averages the first nearest neighbor.
    
    See Also: REGCORR, DETREND_TTS.
    """

    ## Parameters and Initialization
    Nm = np.shape(data)[0]
    Nt = np.shape(data)[1]
    cs = np.unique(info['pairs']['WL'])
    Nc = len(cs)
    hem = np.zeros(shape = [Nc,Nt])

    if sel_type == 'r2d':
        keep_R_NN = np.logical_and((info['pairs']['r2d'] >= value[0]), info['pairs']['r2d'] <= value[1]).astype(np.uint8)
    elif sel_type == 'r3d':
        keep_R_NN = np.logical_and((info['pairs']['r3d'] >= value[0]), info['pairs']['r3d'] <= value[1]).astype(np.uint8)
    elif sel_type == 'NN':
        keep_R_NN = np.zeros(shape = info['pairs']['NN'])
        k = value
        while k == value:
            keep_R_NN = np.logical_or(keep_R_NN, (info['pairs']['NN'] == k)).astype(np.uint8)
            k = k+1

    if np.logical_and(('MEAS' in info), (not 'GI' in info['MEAS'])):
        info['MEAS']['GI'] = np.ones(shape = (Nm, 1), dtype = np.bool8)
    elif not 'MEAS' in info:
        info['MEAS'] = { 'GI': np.ones(shape = (Nm, 1), dtype = np.bool8)}

    k = 0
    while k <=1:
        keep = np.logical_and(np.logical_and(keep_R_NN, (info['pairs']['WL'] == cs[k])), info['MEAS']['GI']).astype(np.uint8)
        hem[k, :] = np.mean(data[np.squeeze(np.argwhere(keep == 1)), :], 0) 
        k = k + 1
    return hem

def highpass(data_in, omegaHz, frate, params = None):
    """ 
    HIGHPASS Applies a zero-phase digital highpass filter.
    data_out = HIGHPASS(data_in, omegaHz, frate) takes a light-level array
    "data_in" in the MEAS x TIME format and applies to it a
    forward-backward zero-phase digital highpass filter at a Nyquist cutoff
    frequency of "omegaHz * (2 * frate)", returning it as "data_out".

    This function also removes the linear component of the input data.

    See Also: LOWPASS, LOGMEAN, FILTFILT.
    """

    ## Parameters and Initialization.
    if params == None:
        params = {}

    if 'poles' in params:
        poles = params['poles']
    else:
        poles = 5
    if 'pad' in params:
        Npad = int(params['pad'])
    else:
        Npad = 100
    if 'detrend' in params:
        DoDetrend = params['detrend']
    else:
        DoDetrend = 1
    if 'DoPad' in params:
        DoPad = params['DoPad']
    else:
        DoPad = 1

    if not DoPad:
        Npad = 0

    if isinstance(data_in, np.single):
        isSing = 1
    else:
        isSing = 0

    dims = np.shape(data_in)
    Nt = dims[-1]  # Assumes time is always the last dimension.
    NDtf = np.ndim(data_in) > 2

    ## N-D Input.
    if NDtf:
        data_in = np.reshape(data_in, len(data_in)/Nt, Nt)
    
    Nm=np.shape(data_in)[0]

    ## Calculate Nyquist frequency and build filter.
    omegaNy = omegaHz * (2 / frate)
    [b, a] = sig.butter(poles, omegaNy, 'highpass')
    
    ## Remove mean
    if data_in.ndim == 1: # When data_in is a vector
        meanRow = np.ndarray.mean(data_in, dtype = np.float64)
        data_in= data_in - meanRow
    else:
        meanRow = np.ndarray.mean(data_in, axis = 1,dtype = np.float64)
        data_in= data_in - meanRow[:,None]

    ## Detrend.
    if DoDetrend:
        data_in = np.squeeze(sig.detrend(data_in[:,None])) 

    ## Zero pad    
    array = np.zeros((int(Nm),int(Npad)))
    data_in = np.append(array,data_in,axis = 1)
    data_in = np.append(data_in,array,axis = 1)

    ## Forward-backward filter data for each measurement.
    if isSing:
        data_in = np.double(data_in)
    data_out = sig.filtfilt(b, a, data_in, padlen = 3*(max(len(b), len(a)) - 1))

    if isSing:
        data_out = np.single(data_out)
    data_out = data_out[:,(Npad):(-Npad)] 

    ## Remove mean
    meanRow = np.ndarray.mean(data_out, axis = 1,dtype = np.float64)
    data_out = data_out - meanRow[:,None]
    ## N-D Output.
    if NDtf:
        data_out = np.reshape(data_out, dims)
    
    return data_out

def logmean(data_in):
    """
    LOGMEAN Computes the log-ratio of raw intensity data.
    
    data_out = LOGMEAN(data_in) takes a light-level data array "data_in" of
    the format MEAS x TIME, and takes the negative log of each element of a
    row divided by that row's average. The result is output into "data_out"
    in the same MEAS x TIME format.
    
    The formal equation for the LOGMEAN operation is:
        Y_{out} = -log(phi_{in} / <phi_{in}>)
    
    If the raw optical data phi is complex (as in the frequency domain 
    case), Y behaves a bit differently. Phi can be defined in terms of Real
    and Imaginary parts: Phi = Re(Phi) + 1i.*Im(Phi), or in terms of it's
    magnitude (A) and phase (theta): Phi = A.*exp(1i.*theta).
    The temporal average of Phi (what we use for baseline) is best
    calculated on the Real/Imaginary decription: 
    
        Phi_o=<phi>=mean(data_in,2) = A_o*exp(i*(th_o));
    
    Taking the logarithm of complex ratio:
    
        Y_Rytov=-log(phi/<phi>)=-log[A*exp(i*th)/A_o*exp(i*th_o)]
                                =-[log(A/A_o) + i(th-th_o)];
    
        Y_Rytov_Re=-log(abs(data_in/Phi_o));
        Y_Rytov_Im=-angle(data_in/Phi_o);
    
    Though this looks like 1 complex number, these components of Y should
    not mix, so the imaginary component will be tacked onto the end of the
    measurement list to keep them separate.
    
    
    Example: If data = [1, 10, 100; exp(1), 10*exp(1), 100*exp(1)];
    
    then LOGMEAN(data) yields [3.6109, 1.3083, -.9943; 3.6109, 1.3083,
    -.9943].
    
    See Also: LOWPASS, HIGHPASS.
    """

    # Parameters and Initialization.
    dims = np.array(np.shape(data_in))
    Nt = dims[-1] # Assumes time is always the last dimension
    NDtf = np.ndim(data_in) > 2
    isZ = not np.any(np.isreal(data_in)) #check if there is at least one element in data_in that is a complex number
    # N-D Input.
    if NDtf:
        data_in = np.reshape(data_in, [], Nt)

    # Perform Logmean.
    Phi_0 = np.mean(data_in,1)
    X = data_in / Phi_0[:,None]

    if not isZ:
        data_out = np.double(-np.log(X))
    else:
        Y_Rytov_Re = -np.log(abs(X))
        Y_Rytov_Im = -np.angle(X)

        data_out = np.concatenate((Y_Rytov_Re, Y_Rytov_Im), axis = 0) # Concatenate along first dimension
        dims[0] = 2*dims[0]

    # Fix any NaNs.
    data_out[not data_in[:].any()]=0

    # N-D Output.
    if NDtf:
        data_out = np.reshape(data_out, dims)

    # Return output
    return (data_out, Phi_0)    

def lowpass(data_in, omegaHz,frate,params = None):
    """
    LOWPASS Applies a zero-phase digital lowpass filter.
    
    data_out = LOWPASS(data_in, omegaHz, frate) takes a light-level array
    "data_in" in the MEAS x TIME format and applies to it a
    forward-backward zero-phase digital lowpass filter at a Nyquist cutoff
    frequency of "omegaHz * (2 * frate)", returning it as "data_out".
    
    This function also removes the linear component of the input data.
    
    See Also: HIGHPASS, LOGMEAN.
    """

    ## Parameters and Initialization.
    if params == None:
        params = {}

    if 'poles' in params:
        poles = params['poles']
    else:
        poles = 5
    if 'pad' in params:
        Npad = int(params['pad'])
    else:
        Npad = 100
    if 'detrend' in params:
        DoDetrend = params['detrend']
    else:
        DoDetrend = 1
    if 'DoPad' in params:
        DoPad = params['DoPad']
    else:
        DoPad = 1

    if not DoPad:
        Npad = 0

    if isinstance(data_in, np.single):
        isSing = 1
    else:
        isSing = 0

    dims = np.shape(data_in)
    Nt = dims[-1]  # Assumes time is always the last dimension.
    NDtf = np.ndim(data_in) > 2

    ## N-D Input.
    if NDtf:
        data_in = np.reshape(data_in, len(data_in)/Nt, Nt)

    Nm=np.shape(data_in)[0]

    ## Calculate Nyquist frequency and build filter.
    omegaNy = omegaHz * (2 / frate)
    [b, a] = sig.butter(poles, omegaNy, 'lowpass')
    
    ## Remove mean
    if data_in.ndim == 1: # When data_in is a vector
        meanRow = np.ndarray.mean(data_in, dtype = np.float64)
        data_in= data_in - meanRow
    else:
        meanRow = np.ndarray.mean(data_in, axis = 1,dtype = np.float64)
        data_in= data_in - meanRow[:,None] 

    ## Detrend.
    if DoDetrend:
        data_in = np.squeeze(sig.detrend(data_in[:,None], type = 'linear'))


    ## Zero pad    
    array = np.zeros((int(Nm),int(Npad)))
    data_in = np.append(array,data_in,axis = 1)
    data_in = np.append(data_in,array,axis = 1)

    ## Forward-backward filter data for each measurement.
    if isSing:
        data_in = np.double(data_in)
    data_out = sig.filtfilt(b, a, data_in, padlen = 3*(max(len(b), len(a)) - 1))

    if isSing:
        data_out = np.single(data_out)
    data_out = data_out[:,(Npad):(-Npad)] 

    ## Detrend.
    if DoDetrend:
        data_out = np.squeeze(sig.detrend(data_out[:,None], type = 'linear'))

    ## Remove mean
    if data_in.ndim == 1: # When data_in is a vector
        meanRow = np.ndarray.mean(data_out, dtype = np.float64)
        data_out= data_out - meanRow
    else:
        meanRow = np.ndarray.mean(data_out, axis = 1,dtype = np.float64)
        data_out = data_out - meanRow[:,None]
    
    ## N-D Output.
    if NDtf:
        data_out = np.reshape(data_out, dims)
    

    return data_out

def regcorr(data_in, info, hem):
    
    """
    REGCORR Performs regression correction by wavelength.

    [data_out, R] = REGCORR(data_in, info, hem) takes a light-level data
    array "data_in" of the format MEAS x TIME, and using the scan metadata
    in "info.pairs" and a WL x MEAS "hem" array generated by GETHEM,
    performs a regression correction for each wavelength of the data, which
    is returned in the MEAS x TIME array "data_out". The corresponding
    correlation coefficients for each measurement are returned in "R" as a
    MEAS x 1 array.

    If y_{r} is the signal to be regressed out and y_{in} is a data
    time trace (either source-detector or imaged), then the output
    is the least-squares regression:

        y_{out} = y_{in} - y_{in}(<y_{in},y_{r}>/|y_{r}|^2). 

    Additionally, the correlation coefficient is given by:
    
        R = (<y_{in},y_{r}>/(|y_{in}|*|y_{r}|)).

    See Also: GETHEM, DETREND_TTS.
    """

    ## Parameters and Initialization.
    Nm = np.shape(data_in)[0]
    Nt = np.shape(data_in)[1]
    cs = np.unique(info['pairs']['WL']) # WLs
    Nc = len(cs)                        # Number of WLs
    data_out = np.zeros(shape = (Nm, Nt))
    R = np.zeros(shape = (Nm, 1))

    ## Regression correction.
    for k in range(0, Nc):
        keep = info['pairs']['WL'] == cs[k]
        temp = np.transpose(data_in[keep, :])

        g = np.transpose(hem[k, :])     # Regressor/noise signal in correct orientation
        g.shape = (g.shape[0], 1)
        gp = lna.pinv(g)
        beta = gp.dot(temp)
        data_out[keep, :] = np.transpose(temp - g.dot(beta)) # Linear regression

        R[keep] = np.transpose(ndot.normrND(np.transpose(g)).dot(ndot.normcND(temp))) # Correlation coefficient
    
    return data_out, R

def resample_tts(data_in, info_in, omega_resample = 1, tol = 0.001, framerate = 0): 
    """
    RESAMPLE_TTS Resamples data while maintaining linear signal components.

    [data_out, info_out] = RESAMPLE_TTS(data_in, info_in, tHz, tol,
    framerate) takes a raw light-level data array "data_in" of the format
    MEAS x TIME, and resamples it (typically downward) to a new frequency
    using the built-in MATLAB function RESAMPLE. The new sampling frequency
    is calculated as the ratio of input "omega_resample" divided by
    "framerate" (both scalars), to within the tolerance specified by "tol".

    This function is needed because the linear signal components, which can
    be important in other NeuroDOT pipeline calculations, can be
    inadvertently removed by downsampling using RESAMPLE alone.

    Note: This function resamples synch points in addition to data. Be sure
    to take care that your data and synch points match after running this
    function! "info.paradigm.init_synchpts" stores the original synch
    points if you need to restore them.

    See Also: DETREND_TTS, RESAMPLE.
    """

    ## Parameters and Initialization.
    info_out = copy.deepcopy(info_in)
    dims = np.shape(data_in)
    Nt = dims[-1]
    NDtf = np.ndim(data_in) > 2
    if framerate == 0:
        if np.logical_and(np.logical_and('system' in info_in, info_in['system'] != {}), np.logical_and('framerate' in info_in['system'], info_in['system']['framerate'] != {})):
            framerate = info_in['system']['framerate'] 

    ## N-D Input.
    if NDtf:
        data_in = np.reshape(data_in, [], Nt)


    ## Approximate desired resampling ratio as a fraction.
    fract = sym.nsimplify((omega_resample/framerate), tolerance = tol, rational = True) # In Matlab, tolerance is set to 1e-5, in order to reproduce outputs in Python, tolerance is set to 0.001 
    N = fract.numerator
    D = fract.denominator
    info_out['system']['framerate'] = omega_resample


    ## Remove linear fit.
    Nt = np.shape(data_in)[1]
    d0 = data_in[:, 0]    # Start point
    dF = data_in[:, Nt-1] # End point
    beta = -d0

    alpha1 = np.divide((d0-dF), (Nt-1)) # Slope for linear fit
    cols = np.arange(0, Nt, 1)          # bsxfun (Matlab) multiplication
    alpha_full = np.ones((dims[0], Nt))
    for x in cols:
        alpha_full[:, x] = cols[x] * alpha1
    beta.shape = (np.shape(beta)[0],1)
    correction = alpha_full + beta      # bsxfun (Matlab) addition
    corrsig = data_in + correction


    ## Resample with endpoints pinned to zero.
    try:
        print(N,D)
        rawresamp = np.transpose(sig.resample_poly(np.transpose(corrsig[:]), N, D)) # Using scipy signal polyphase resampling

    except TypeError:
        N = fract.numerator()
        D = fract.denominator()
        rawresamp = np.transpose(sig.resample_poly(np.transpose(corrsig[:]), N, D)) # Using scipy signal polyphase resampling

    ## Add linear fit back to resampled data.
    alpha2 = alpha1 * (D/N) 
    Nt = np.shape(rawresamp)[1]     # Column dimension
    cols = np.arange(0, Nt, 1)      # bsxfun (Matlab) multiplication
    alpha_full = np.ones((dims[0], Nt))
    for x in cols:
        alpha_full[:, x] = cols[x] * alpha2
    correction = alpha_full + beta  # bsxfun (Matlab) addition
    data_out = rawresamp - correction


    ## Fix synch pts to new framerate.
    if 'paradigm' in info_in:
        if 'init_synchpts' in info_in['paradigm']:
            info_out['paradigm']['synchpts'] = np.round(np.divide(np.dot(N, info_out['paradigm']['init_synchpts']), D)) 
            info_out['paradigm']['synchpts'][np.argwhere(info_out['paradigm']['synchpts'] == 0)] = 1
        elif 'synchpts' in info_in['paradigm']:
            info_out['paradigm']['init_synchpts'] = info_out['paradigm']['synchpts']
            info_out['paradigm']['synchpts'] = np.round(np.divide(np.dot(N, info_out['paradigm']['synchpts']), D))
            info_out['paradigm']['synchpts'][np.argwhere(info_out['paradigm']['synchpts'] == 0)] = 1


    ## N-D Output.
    if NDtf:
        data_out = np.reshape(data_out, (dims[0:-1], Nt))

    return data_out, info_out




# General imports
import numpy as np
import numpy.linalg as lna

import neuro_dot as ndot


   
def BlockAverage(data_in, pulse, dt, Tkeep = 0):
    """ 
    BLOCKAVERAGE Averages data by stimulus blocks.

    data_out = BLOCKAVERAGE(data_in, pulse, dt) takes a data array "data_in" 
    and uses the pulse and dt information to cut that data timewise into 
    blocks of equal length (dt), which are then averaged together and 
    output as "data_out".
    
    Tkeep is a temporal mask. Any time points with a zero in this vector is
    set to NaN.
    """
    ## Parameters and Initialization.
    dims = np.shape(data_in)
    Nt = dims[-1]
    NDtf = np.ndim(data_in) > 2
    Nbl = len(pulse)

    if Tkeep == 0:
        Tkeep = np.ones(shape = (Nt, 1))==1 

    # Check to make sure that the block after the last synch point for this
    if (dt + pulse[-1] - 1) > Nt:
        Nbl = Nbl - 1

    ## N-D Input (for 3-D or N-D voxel spaces).
    if NDtf:
        data_in = np.reshape(data_in, [], Nt)

    ## Incorporate Tkeep  
    data_in[:, np.argwhere(Tkeep == False)] = np.NaN

    ## Cut data into blocks.
    Nm = np.shape(data_in)[0]
    blocks = np.zeros((Nm, dt, Nbl))

    for p in range(0,len(pulse)):
        pulse[p] = pulse[p] +1

    for k in range(0, Nbl):
        pulse_k = int(pulse[k])

        if (pulse[k] + dt -1) <= Nt:
            blocks[:, :, k] = data_in[:, pulse_k-1:pulse_k + dt-1] #Need to subtract 1 from both indices to account for 0 indexing in python
        else:
            dtb = (pulse_k-1) + dt - 1 - Nt+1
            nans = np.empty(shape = (np.shape(data_in)[0], dtb)) # need multiple lines to create an array of nans in python
            nans[:] = np.NaN
            blocks[:, :, k] = np.concatenate((data_in[:, (pulse_k-1):Nt], nans), axis = 1) # need to subtract 1 from start index: pulse[k] due to zero indexing, also need to use Nt as final index to get correct size

    ## Average blocks and return.
    BA_out = np.nanmean(blocks, axis = 2) 
    BSTD_out = np.nanstd(blocks, axis = 2, ddof = 1) 
    nanmean_cols = np.nanmean(BA_out, axis = 1)
    nanmean_matrix = np.ones(np.shape(BA_out))
    for x in range(0, dt): 
        nanmean_matrix[:, x] = nanmean_cols
    BA_out = BA_out - nanmean_matrix
    BT_out = np.divide(BA_out, BSTD_out)
    BT_out[np.argwhere(np.isinf(BT_out))] = 0

    ## N-D Output.
    if NDtf:
        #create tuple containing the desired output shape for BA_out, BSTD_out and BT_out
        #tuple contains first axis until second to last axis of data with dt appeneded to the end of it
        newshape = tuple(np.append(np.array(dims[0:-1]), dt))
        BA_out = np.reshape(BA_out, newshape)
        BSTD_out = np.reshape(BSTD_out, newshape)
        BT_out = np.reshape(BT_out, newshape)
        newshape_blocks = tuple(np.append(np.array(dims[0:-1]), (dt, Nbl))) # create tuple for deisred output shape for blocks, different from previous newshape bc Nbl is also appended
        blocks = np.reshape(blocks, newshape_blocks)


    return BA_out, BSTD_out, BT_out, blocks


def CalcGVTD(data):
    """
    CalcGVTD calculates the Root Mean Square across measurements (log-mean light levels or voxels) of the temporal derivative. 

    The data is assumed to have measurements in the first and time in the last 
    dimension. 
    
    Any selection of measurement type or voxel index must be done
    outside of this function.
    """
    # Double check data has correct dimensions
    # Dsizes=size(data);
    Dsizes = np.shape(data)
    Ndim = len(Dsizes)
    if Ndim > 2:
        data = np.reshape(data, [], Dsizes[-1])
    
    # 1st Temporal Derivative
    Ddata = data - np.roll(data, np.array([0, -1]), np.array([0,-1])) 

    # RMS across measurements
    GVTD = np.concatenate(([0], ndot.rms_py(Ddata[:,0:-1])), axis = 0)

    return GVTD


def FindGoodMeas(data, info_in, bthresh = 0.075):

    """ 
    FINDGOODMEAS Performs "Good Measurements" analysis to return indices of measurements within a chosen threshold.

    info_out = FINDGOODMEAS(data, info_in) takes a light-level array "data"
    in the MEAS x TIME format, and calculates the std of each channel
    as its noise level. 

    These are then thresholded by the default value of
    0.075 to create a logical array, and both are returned as MEAS x 1
    columns of the "info.MEAS" table. 
    
    If pulse synch point information exists in "info.system.synchpts",
    then FINDGOODMEAS will crop the data to the start and stop pulses.

    info_out = FINDGOODMEAS(data, info_in, bthresh) allows the user to
    specify a threshold value.
    
    See Also: PLOTCAPGOODMEAS, PLOTHISTOGRAMSTD.
    """
    ## Parameters and Initialization.
    # look for required info, FGM will not run if these fields are nonexistant
    try:
        info1 = info_in['system']['framerate']
    except KeyError:
        print('info_in["system"]["framerate"] does not exist and is required')
        print('exiting FindGoodMeas')
        return()
    try:
        info1 = info_in['pairs']
    except KeyError:
        print('info_in["pairs"] does not exist and is required')
        print('exiting FindGoodMeas')
        return()
    info_out = info_in.copy() # create info_out
    try:
        GVwin
    except NameError:
        GVwin = 600
    if not 'paradigm' in info_out:
        info_out['paradigm'] = {}
    if not bthresh in locals():
        bthresh = 0.075 # Empirically derived threshold value.
    dims = data.shape
    Nt = dims[-1]       # Assumes time is always the last dimension
    NDtf = np.ndim(data) > 2
    if GVwin > (Nt-1):
        GVwin = (Nt-1)
    
    # N-D Input.
    if NDtf:
        data = np.reshape(data, [], Nt)
    
    # Crop data to synchpts if necessary. 
    keep = np.logical_and(info_in['pairs']['r2d'] < 20, info_in['pairs']['WL'] == 2)
    foo = np.squeeze(data[keep,:])
    foo = ndot.highpass(foo, 0.02, info_in['system']['framerate']) # bandpass filter, omega_hp = 0.02
    foo = ndot.lowpass(foo, 1, info_in['system']['framerate'])     # bandpass filter, omega_lp = 1
    foo = foo - np.roll(foo, 1, 1)
    foo[:,0] = 0
    foob = ndot.rms_py(foo) # uses new rms that only takes one input, calculates rms for every column in rms_input and outputs row vector
    NtGV = Nt - GVwin
    NtGV_mat = np.ones((1,NtGV), dtype = np.int8)  

    if NtGV > 1: # sliding window to grab a meaningful set for 'quiet'
        GVTD_win_means = np.zeros(NtGV, order = 'F')
        i = 0
        while i <= (NtGV - 1):
            GVTD_win_means[i] =  np.mean(foob[i:((i+1)+ GVwin - 1)])
            i = i+1
        t0 = np.where(GVTD_win_means == np.min(GVTD_win_means)) # find min and set t0 --> tF
        tF = t0[0][0] + GVwin 
        STD = np.std(data[:, t0[0][0]:tF], 1, ddof= 1)          # Calulate STD, make sure ddof param is set = 1 so that np.STD behaves the same as matlab STD
    elif not 'synchpts' in info_out['paradigm']:
        NsynchPts = len(info_out['paradigm']['synchpts'])
        if NsynchPts > 2:
            tF = info_out['paradigm']['synchpts'][-1]
            t0 = info_out['paradigm']['synchpts'][0]
        elif NsynchPts == 2:
            tF = info_out['paradigm']['synchpts'][1]
            t0 = info_out['paradigm']['synchpts'][0]
        else:
            tF = data.shape[1]
        STD = np.std(data[:, t0:tF], 1, ddof=1)                 # Calculate STD.
    else:
        STD = np.std(data, 1, ddof=1)
    
    # Populate in table of on-the-fly calculated stuff.
    info_out['GVTDparams'] = {}
    info_out['GVTDparams']['t0'] = t0[0][0]
    info_out['GVTDparams']['tF'] = tF
    if not 'MEAS' in info_out:
        info_out['MEAS'] = {}
        info_out['MEAS']['STD'] = STD
        info_out['MEAS']['GI'] = np.zeros(np.shape(STD), dtype = np.uint8)
        info_out['MEAS']['GI'][np.where(STD <= bthresh)] = 1
    else:
        info_out['MEAS']['STD'] = STD
        info_out['MEAS']['GI'] = np.zeros(np.shape(STD), dtype = np.uint8)
        info_out['MEAS']['GI'][np.where(STD <= bthresh)] = 1
    if 'Clipped' in info_out['MEAS']:
        info_out['MEAS']['GI'] = np.zeros(np.shape(STD), dtype = np.uint8)
        info_out['MEAS']['GI'][np.where(info_out['MEAS']['GI'] and not info_out['MEAS']['Clipped'])] = 1
    
    return info_out


def normcND(data):
    """ 
    NORMCND returns a column-normed matrix. It is assumed that the matrix is 2D.
    """
    vecnorm = lna.norm(data, ord = 2, axis = 0)
    data = data / vecnorm

    return data


def normrND(data):
    """
    NORMRND returns a row-normed matrix. It is assumed that the matrix is 2D. Updated for broader compatability.
    """
    dataNorm = np.sqrt(np.sum(data**2))
    data = data / dataNorm
    data[np.argwhere(np.isfinite(data) == False)] = 0

    return data
